# -*- coding: utf-8 -*-
'''
Created on 6 мая 2020 г.

@author: dvd
'''
from __future__ import unicode_literals

from django.core.management.base import BaseCommand, CommandError
from django.utils import translation
from uds.models import Authenticator
from uds.core import auths

import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

import argparse
import re


def small_name_type(val):
    max_length = Authenticator._meta.get_field('small_name').max_length
    if len(val.decode('utf8')) > max_length:
        error = "Length exceeds {} chars".format(max_length)
        raise argparse.ArgumentTypeError(error)

    if re.search("(?u)[^ a-zA-Z0-9_]", val):
        error = "Allowed chars: a-z A-Z 0-9 _".format(max_length)
        raise argparse.ArgumentTypeError(error)

    return val


def positive_int_type(val):
    ivalue = int(val)
    if ivalue <= 0:
        error = "Positive value required: {}".format(val)
        raise argparse.ArgumentTypeError(error)
    return ivalue


class Command(BaseCommand):
    help = '''
            Создает домен аутентификации.
            Для просмотра параметров домена аутентификации 
            не указывайте аргумент --params.
            '''

    def add_arguments(self, parser):
        parser.add_argument(
            '--name', type=str, required=True,
            help="Имя домена аутентификации"
        )

        parser.add_argument(
            '--small_name', type=small_name_type, required=True,
            help=("Короткое имя домена аутентификации. "
                  "Допустимые символы: a-z A-Z 0-9 _. "
                  "Максимум 32 символа.")
        )

        parser.add_argument(
            '--priority', type=positive_int_type,
            help="Приоритет домена аутентификации. Значение по-умолчанию: 1.",
            default=1
        )

        parser.add_argument(
            '--comments', type=str, help="Комментарий", default=''
        )
        # 2to3: auths.factory().providers().values() passed to list()
        auth_types = [p.typeType for p in
                      list(auths.factory().providers().values())]
        parser.add_argument(
            '--type', choices=auth_types, required=True,
            help='Тип домена аутентификации'
        )

        parser.add_argument(
            '--params', nargs='*', type=str,
            help=('Параметры домена аутентификации. '
                  'Формат: имя1=значeние1 имя2=значение2 ...'),
            default=[]
        )

        parser.add_argument(
            '--list', action='store_true', dest='list_params', default=False,
            help='Показать параметры домена аутентификации',
        )

        parser.add_argument(
            '--test', action='store_true', dest='test', default=False,
            help='Проверка параметров без создания домена аутентификации',
        )

        parser.add_argument(
            '--quiet', action='store_true', dest='quiet', default=False,
            help='Выводить только сообщения об ошибках',
        )

    def handle(self, *args, **options):

        auth_type = options['type']

        if options['list_params']:
            self.list_auth_params(auth_type)
            return

        base_args = ['name', 'small_name', 'priority', 'comments']
        base_data = {k: options[k] for k in base_args}
        self.validate_base_data(base_data)

        params = self.parse_params(options)
        test_success, test_msg = self.test(auth_type, params)
        if test_success:
            if options['test']:
                if not options['quiet']:
                    self.stdout.write(
                        self.style.SUCCESS('Проверка прошла успешно')
                    )
                return
        else:
            self.stderr.write(self.style.ERROR(test_msg))
            return

        item = Authenticator.objects.create(**base_data)
        item.data_type = auth_type
        item.data = item.getInstance(params).serialize()
        item.save()

        if not options['quiet']:
            self.stdout.write(
                self.style.SUCCESS('Домен аутентификации успешно создан.')
            )

    def test(self, type_, data):
        from uds.core.Environment import Environment

        authType = auths.factory().lookup(type_)

        res = authType.test(Environment.getTempEnv(), data)
        return res

    def list_auth_params(self, auth_type):
        gui = auths.factory().lookup(auth_type).guiDescription()
        header = "Тип {} поддерживает следующие параметры:".format(auth_type)
        self.stdout.write(
            self.style.SUCCESS(header)
        )
        for param in gui:
            self.stdout.write(
                self.style.WARNING("    {}".format(param['name']))
            )

            tooltip = param['gui']['tooltip']
            self.stdout.write("        {}".format(tooltip))

            defval = param['gui']['defvalue']
            if defval:
                defval_str = "        По-умолчанию: {}".format(defval)
                self.stdout.write(
                    self.style.SUCCESS(defval_str)
                )

            if param['gui']['required']:
                self.stdout.write(self.style.ERROR("        Обязательный"))

            self.stdout.write(self.style.ERROR(""))

    def parse_params(self, options):
        data = {}
        for field in options['params']:
            field = field.decode('utf-8')
            pair = field.split('=')
            if len(pair) == 2:
                name, value = pair
            else:
                name, value = pair[0], '='.join(pair[1:])

            data[name] = value
        return data

    def validate_base_data(self, data):
        # Проверка наличия в БД
        try:
            Authenticator.objects.get(name=data['name'])
            error = ('Домен аутентификации {} '
                     'уже существует').format(data['name'])
            raise CommandError(error)
        except Authenticator.DoesNotExist:
            pass

        try:
            Authenticator.objects.get(small_name=data['small_name'])
            error = ('Домен аутентификации c коротким именем {} '
                     'уже существует').format(data['small_name'])
            raise CommandError(error)
        except Authenticator.DoesNotExist:
            pass